@@ -104,6 +104,8 @@ struct QuantileConsts {
       {.5, .75, .90, .95, .99}};
   static constexpr const std::array<double, 5> kP1_P10_P50_P90_P99{
       {.01, .1, .5, .9, .99}};
+  static constexpr const std::array<double, 6> kP1_P10_P75_P90_P95_P99{
+      {.01, .1, .75, .9, .95, .99}};
   static constexpr const std::array<double, 6> kP50_P75_P90_P95_P99_P999{
       {.5, .75, .9, .95, .99, .999}};
   static constexpr const std::array<double, 5> kP50_P70_P90_P95_P99{
