@@ -16,10 +16,11 @@
 
 #include <fb303/ServiceData.h>
 
-#include <boost/regex.hpp>
 #include <stdexcept>
 
+#include <boost/regex.hpp>
 #include <fb303/LegacyClock.h>
+#include <fb303/detail/RegexUtil.h>
 #include <folly/Conv.h>
 #include <folly/ExceptionString.h>
 #include <folly/Indestructible.h>
@@ -40,7 +41,6 @@ template <typename T>
 static T& as_mutable(T const& t) {
   return const_cast<T&>(t);
 }
-
 /*
  * The constructor used to create additional ServiceData instances.
  * IMPORTANT NOTE: There already is a global singleton instance living,
@@ -78,9 +78,14 @@ void ServiceData::flushAllData() {
 
 void ServiceData::resetAllData() {
   options_.wlock()->clear();
+  {
+    auto countersWLock = counters_.wlock();
+    countersWLock->map.clear();
 
-  counters_.wlock()->clear();
-
+    // avoid fetch_add() to avoid extra fences, since we hold the lock already
+    uint64_t epoch = countersWLock->mapEpoch.load();
+    countersWLock->mapEpoch.store(epoch + 1);
+  }
   exportedValues_.wlock()->clear();
 
   statsMap_.forgetAllStats();
@@ -92,12 +97,13 @@ void ServiceData::resetAllData() {
 }
 
 void ServiceData::zeroStats() {
-  counters_.withRLock([&](auto const& counters) {
-    for (auto const& elem : counters) {
+  {
+    auto countersRLock = counters_.rlock();
+    for (auto const& elem : countersRLock->map) {
       //  this const-cast is safe: the lock protects the map structure only
       as_mutable(elem.second).store(0, std::memory_order_relaxed);
     }
-  });
+  }
 
   statsMap_.clearAllStats();
   histMap_.clearAllHistograms();
@@ -289,7 +295,7 @@ int64_t ServiceData::incrementCounter(StringPiece key, int64_t amount) {
   {
     //  optimistically, the key is certainly present; update under rlock
     auto countersRLock = counters_.rlock();
-    if (auto ptr = folly::get_ptr(*countersRLock, key)) {
+    if (auto ptr = folly::get_ptr(countersRLock->map, key)) {
       //  this const-cast is safe: the lock protects the map structure only
       auto& ref = as_mutable(*ptr);
       return ref.fetch_add(amount, std::memory_order_relaxed) + amount;
@@ -298,7 +304,12 @@ int64_t ServiceData::incrementCounter(StringPiece key, int64_t amount) {
 
   //  pessimistically, the key is possibly absent; upsert under wlock
   auto countersWLock = counters_.wlock();
-  auto& ref = (*countersWLock)[key];
+  auto& ref = (countersWLock->map)[key];
+
+  // avoid fetch_add() to avoid extra fences, since we hold the lock already
+  uint64_t epoch = countersWLock->mapEpoch.load();
+  countersWLock->mapEpoch.store(epoch + 1);
+
   return ref.fetch_add(amount, std::memory_order_relaxed) + amount;
 }
 
@@ -306,7 +317,7 @@ int64_t ServiceData::setCounter(StringPiece key, int64_t value) {
   {
     //  optimistically, the key is certainly present; update under rlock
     auto countersRLock = counters_.rlock();
-    if (auto ptr = folly::get_ptr(*countersRLock, key)) {
+    if (auto ptr = folly::get_ptr(countersRLock->map, key)) {
       //  this const-cast is safe: the lock protects the map structure only
       auto& ref = as_mutable(*ptr);
       ref.store(value, std::memory_order_relaxed);
@@ -316,16 +327,24 @@ int64_t ServiceData::setCounter(StringPiece key, int64_t value) {
 
   //  pessimistically, the key is possibly absent; upsert under wlock
   auto countersWLock = counters_.wlock();
-  auto& ref = (*countersWLock)[key];
+  auto& ref = (countersWLock->map)[key];
+
+  // avoid fetch_add() to avoid extra fences, since we hold the lock already
+  uint64_t epoch = countersWLock->mapEpoch.load();
+  countersWLock->mapEpoch.store(epoch + 1);
+
   ref.store(value, std::memory_order_relaxed);
   return value;
 }
 
 void ServiceData::clearCounter(StringPiece key) {
   auto countersWLock = counters_.wlock();
-  auto it = countersWLock->find(key);
-  if (it != countersWLock->end()) {
-    countersWLock->erase(it);
+  auto it = countersWLock->map.find(key);
+  if (it != countersWLock->map.end()) {
+    // avoid fetch_add() to avoid extra fences, since we hold the lock already
+    uint64_t epoch = countersWLock->mapEpoch.load();
+    countersWLock->mapEpoch.store(epoch + 1);
+    countersWLock->map.erase(it);
   }
 }
 
@@ -342,7 +361,7 @@ folly::Optional<int64_t> ServiceData::getCounterIfExists(
   }
 
   auto countersRLock = counters_.rlock();
-  auto ptr = folly::get_ptr(*countersRLock, key);
+  auto ptr = folly::get_ptr(countersRLock->map, key);
   return ptr ? folly::make_optional(ptr->load(std::memory_order_relaxed))
              : folly::none;
 }
@@ -358,25 +377,31 @@ int64_t ServiceData::getCounter(StringPiece key) const {
 }
 
 void ServiceData::getCounters(map<string, int64_t>& _return) const {
-  counters_.withRLock([&](auto const& counters) {
-    for (auto const& elem : counters) {
-      _return[elem.first] = elem.second.load(std::memory_order_relaxed);
+  {
+    auto countersRLock = counters_.rlock();
+    for (auto const& elem : countersRLock->map) {
+      _return.insert(std::make_pair(
+          std::string(elem.first),
+          elem.second.load(std::memory_order_relaxed)));
     }
-  });
+  }
 
   quantileMap_.getValues(_return);
 
   dynamicCounters_.getCounters(&_return);
 }
 
+void ServiceData::getKeys(std::vector<std::string>& keys) const {
+  auto countersRLock = counters_.rlock();
+  keys.reserve(keys.size() + countersRLock->map.size());
+  for (const auto& [key, _] : countersRLock->map) {
+    keys.emplace_back(key);
+  }
+}
+
 std::vector<std::string> ServiceData::getCounterKeys() const {
   std::vector<std::string> keys;
-  counters_.withRLock([&](auto const& counters) {
-    keys.reserve(counters.size());
-    for (const auto& elem : counters) {
-      keys.push_back(elem.first);
-    }
-  });
+  getKeys(keys);
 
   quantileMap_.getKeys(keys);
 
@@ -388,8 +413,7 @@ std::vector<std::string> ServiceData::getCounterKeys() const {
 uint64_t ServiceData::getNumCounters() const {
   int64_t numCounters = 0;
 
-  counters_.withRLock(
-      [&](auto const& counters) { numCounters += counters.size(); });
+  numCounters += counters_.rlock()->map.size();
 
   numCounters += quantileMap_.getNumKeys();
 
@@ -405,19 +429,29 @@ map<string, int64_t> ServiceData::getCounters() const {
 }
 
 void ServiceData::getSelectedCounters(
-    map<string, int64_t>& _return,
+    map<string, int64_t>& output,
     const vector<string>& keys) const {
-  quantileMap_.getSelectedValues(_return, keys);
-  for (const string& key : keys) {
-    if (_return.find(key) == _return.end()) {
-      try {
-        int64_t value = getCounter(key);
-        _return[key] = value;
-      } catch (const std::invalid_argument&) {
-        // Don't insert anything into _return for non-existent keys.
+  // lock once and grab all the flat counters in one go...
+  {
+    auto countersRLock = counters_.rlock();
+    for (const string& key : keys) {
+      auto ptr = folly::get_ptr(countersRLock->map, key);
+      if (ptr) {
+        output[key] = ptr->load(std::memory_order_relaxed);
       }
     }
   }
+
+  // dynamic counters can replace flat counters
+  for (const string& key : keys) {
+    int64_t ret;
+    if (dynamicCounters_.getValue(key, &ret)) {
+      output[key] = ret;
+    }
+  }
+
+  // quantiles can replace flat and dynamic counters
+  quantileMap_.getSelectedValues(output, keys);
 }
 
 map<string, int64_t> ServiceData::getSelectedCounters(
@@ -431,7 +465,6 @@ void ServiceData::getRegexCounters(
     map<string, int64_t>& _return,
     const string& regex) const {
   const boost::regex regexObject(regex);
-
   auto keys = getCounterKeys();
   keys.erase(
       std::remove_if(
@@ -445,12 +478,29 @@ void ServiceData::getRegexCounters(
   getSelectedCounters(_return, keys);
 }
 
+void ServiceData::getRegexCountersOptimized(
+    map<string, int64_t>& output,
+    const string& regex) const {
+  std::vector<std::string> keys;
+  detail::getRegexKeysImpl(keys, regex, counters_);
+  quantileMap_.getRegexKeys(keys, regex);
+  dynamicCounters_.getRegexKeys(keys, regex);
+  getSelectedCounters(output, keys);
+}
+
 map<string, int64_t> ServiceData::getRegexCounters(const string& regex) const {
   map<string, int64_t> _return;
   getRegexCounters(_return, regex);
   return _return;
 }
 
+map<string, int64_t> ServiceData::getRegexCountersOptimized(
+    const string& regex) const {
+  map<string, int64_t> output;
+  getRegexCountersOptimized(output, regex);
+  return output;
+}
+
 bool ServiceData::hasCounter(StringPiece key) const {
   if (dynamicCounters_.contains(key)) {
     return true;
@@ -460,7 +510,7 @@ bool ServiceData::hasCounter(StringPiece key) const {
     return true;
   }
 
-  return counters_.rlock()->count(key) != 0;
+  return counters_.rlock()->map.count(key) != 0;
 }
 
 void ServiceData::deleteExportedKey(StringPiece key) {
